CompileOptions

Struct CompileOptions 

pub struct CompileOptions {
    pub ftz: Option<bool>,
    pub prec_sqrt: Option<bool>,
    pub prec_div: Option<bool>,
    pub fmad: Option<bool>,
    pub options: Vec<String>,
    pub use_fast_math: Option<bool>,
    pub maxrregcount: Option<usize>,
    pub include_paths: Vec<String>,
    pub arch: Option<&'static str>,
    pub name: Option<String>,
}
Flags you can pass to the nvrtc compiler. See https://docs.nvidia.com/cuda/nvrtc/index.html#group__options for all available flags and documentation for what they do.

All fields of this struct match one of the flags in the documentation. if a field is None it will not be passed to the compiler.

All fields default to None.

NOTE: not all flags are currently supported.

Example:

// "--ftz=true" will be passed to the compiler
let opts = CompileOptions {
    ftz: Some(true),
    ..Default::default()
};

Fields

ftz: Option<bool>prec_sqrt: Option<bool>prec_div: Option<bool>fmad: Option<bool>options: Vec<String>use_fast_math: Option<bool>maxrregcount: Option<usize>include_paths: Vec<String>arch: Option<&'static str>name: Option<String>