pub struct MappedBuffer { /* private fields */ }Expand description
An abstraction for a mapped buffer for some external memory.
This struct can be created via ExternalMemory::map_range or ExternalMemory::map_all.
The underlying mapped buffer will be freed when this struct is dropped.